//===----------- FernandoLiveStack.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it stores liveness information for the control flow graph.
//
// Date: May 28th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file contains liveness information for the control flow graph.
// For each pair of basic blocks u and v, it records the set of variables
// alive between u and v. The requirements in terms of space is O(N*N*T),
// where N is the number of basic blocks, and T is the total number of
// temporaries in the control flow graph. Notice that this program only
// records information about virtual registers. Physical registers are
// not supposed to be alive between basic blocks.
//
//===----------------------------------------------------------------------===//

#ifndef EDGE_LIVENESS_FERNANDO_H
#define EDGE_LIVENESS_FERNANDO_H

#include "llvm/Analysis/LoopInfo.h"
#include "llvm/fernando/CriticalEdgeRemoval_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/PassSupport.h"

namespace llvm {

    #define UNDEF_REG 65535

    class EdgeLiveness_Fernando : public MachineFunctionPass {

        public:

        /// The ID for this pass.
        const PassInfo *EdgeLivenessFernandoPassID;

        /// This method fills a vector with all the virtual registers alive
        /// between the origin and destiny blocks.
        void to_vector(
            const MachineBasicBlock & origin,
            const MachineBasicBlock & destiny,
            std::vector<unsigned> & ans
        ) const;

        /// returns a string in the format (t1, t2, ..., tn), where ti is
        /// one of the temporaries alive in the edge that connects the
        /// machine basic block "origin" to "destiny".
        std::string to_string(
            const MachineBasicBlock & origin,
            const MachineBasicBlock & destiny
        ) const;

        /// Returns a vector containing the registers alive in the edge
        /// between the given basic blocks. Notice that each element in the
        /// returned vector is not the LLVM representation of the virtual
        /// register. To get the virtual register number, add to the
        /// MRegisterInfo::FirstVirtualRegister to the index in the returned
        /// vector.
        const std::vector<bool> & get_live_variables(
            const MachineBasicBlock & origin,
            const MachineBasicBlock & destiny
        ) const;

        /// reporst that this pass does not change anything on the control
        /// flow graph.
        virtual void getAnalysisUsage(AnalysisUsage &AU) const {
            AU.addRequired<CriticalEdgeRemoval_Fer>();
            AU.addRequired<DefUseSites_Fer>();
            AU.setPreservesAll();
        }

        /// Informs the name of the pass, so it can appear in tools like
        /// llc.
        virtual const char *getPassName() const {
            return "Edge liveness analyses - Fernando.";
        }

        /// This method informs if the given temporary is alive in the edge
        /// between the blocks "origin" and "destiny".
        bool is_alive(
                       const MachineBasicBlock & origin,
                       const MachineBasicBlock & destiny,
                       const unsigned v_reg
                     ) const;

        /// Informs the maximum number of temporary registers that can be
        /// alive in any time in any edge of the control flow graph. This is
        /// the number of temporaries used in the target program.
        unsigned get_num_virtual_registers() const {
            return this->num_vreg;
        }

        /// This method removes the live range of v_sp. This must be done after
        /// spill is performed. The placement of load and store instructions
        /// must be performed in such a way that is always safe to remove the
        /// live range of the spilled register.
        void kill_live_range
                         (const unsigned v_reg, const MachineBasicBlock & mbb);

        /// The method prints all the results obtained during the liveness
        /// analysis. It only print liveness analysis results. Although this
        /// class also contains a data structure that marks the block where
        /// each temporary is defined, the dump method does not print this
        /// information.
        void dump() const;

        private:

        /// The only purpose of keeping the machine function in this class is
        /// to help in the dump method.
        MachineFunction * machine_function;

        /// This variables records the maximum number of virtual registers that
        /// can be alive at any moment in any edge of the control flow graph.
        /// This is the number of temporaries (also called virtual) registers
        /// in the target program.
        unsigned num_vreg;

        /// the data structure where liveness information will be stored.
        std::vector< std::vector < std::vector<bool> > > live_sets;

        /// This object contains all the information about the register set
        /// in the target program, including the number of registers.
        const MRegisterInfo * reg_info;

        /// This variable maps virtual registers to the instructions where it
        /// is defined and used.
        DefUseSites_Fer * def_use_sites;

        /// runOnMachineFunction - collects liveness information about each
        /// edge of the control flow graph.
        bool runOnMachineFunction(MachineFunction &mf);

        /// Removes the v_reg from the live set of all the blocks that can be
        /// reached from mbb.
        void kill_forward(const unsigned v_reg, const MachineBasicBlock & mbb);

        /// Removes the v_reg from the live set of all the blocks that have a
        /// path leading to mbb.
        void kill_backward(const unsigned v_reg,const MachineBasicBlock & mbb);

        /// adds the temporary to the live set of the edge that goes from
        /// machine basic block "origin" to "destiny".
        void set_alive(
                        const MachineBasicBlock & origin,
                        const MachineBasicBlock & destiny,
                        const unsigned v_reg
                      );

        /// Removes temporary from the live set of the edge connecting the
        /// block origin to the block destiny.
        void kill(
                   const MachineBasicBlock & origin,
                   const MachineBasicBlock & destiny,
                   const unsigned v_reg
                 );

        /// This function scans every instruction of the basic block, and
        /// compute liveness information in a backward fashion.
        void compute_liveness(const MachineBasicBlock & block);

        /// This method allocates memory for the matrix.
        void initialize_data_structures(MachineFunction & mf);

        /// Every time a use is found, the graph is scanned backwards, and all
        /// the edges in which the used temporary is alive are marked. This
        /// function does exaclty this.
        void handle_use(
                         const unsigned v_reg,
                         const MachineBasicBlock & block
                        );

        /// Free the memory, so the operating system can use it in other
        /// applications. It deletes the mapping between temporaries to blocks
        /// where they are defined, and the matrix that contains liveness
        /// information.
        void releaseMemory();

    };  // end of EdgeLiveness_Fernando declaration

} // end llvm namespace

#endif
