//===----------- IG_Builder_Fer - Builds the interference graph ----------===//
//
//                     The LLVM Compiler Infrastructure
//
// Given a control flow graph, this pass builds the correcsponding interference
// graph. This file has been implemented by fernando at cs dot ucla dot edu.
//
// Date: July 10th, 2006
//
//===---------------------------------------------------------------------===//
//
// This pass builds the interference graph for a target control flow graph.
// Two nodes in the interference graph are connected if, and only if, their
// live ranges overlap in the control flow graph.
// Important: this version of the interference graph does not contains the
// full interfering pairs. It only holds interferences between physical
// registers and virtual registers. This helps in the implementation of the
// register allocation pass, and avoids using too much memory and time to
// hold the interference graph. Actually, this graph uses about one fourth
// of the memory that it would use if the full interference graph were to be
// built and stored.
//
//===---------------------------------------------------------------------===//


#ifndef IG_BUILDER_FER_H
#define IG_BUILDER_FER_H

#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/RegClique_Fer.h"

#include "llvm/Pass.h"

namespace llvm {

    class IG_Builder_Fer : public MachineFunctionPass {

    public:

    /// reporst that this pass does not change anything on the control
    /// flow graph.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const {
        AU.addRequired<EdgeLiveness_Fernando>();
        AU.setPreservesAll();
    }


    /// The label of this pass is "Interference graph builder."
    virtual const char * getPassName() const {
        return "Interference graph builder.";
    }


    /// This method builds the interference graph between physical and
    /// virtual registers.
    virtual bool runOnMachineFunction(MachineFunction &mf);
 

    /// This method prints the constraints in the standard error output.
    void dump_constraints();


    /// This method modifies the cliques in the cl vector to reflect the newly
    /// created virtual registers. Such virtuals have been created due to
    /// spilling.
    void rebuild_list_cliques(
        unsigned sp,
        std::vector<RegClique_Fer *> &cl,
        const MachineBasicBlock &mbb
    );


    /// If liveness analysis information is available, the set of cliques in a
    /// basic block can be found with a backwards linear scan. The first clique
    /// will be the out set of the basic block. Each time a new instruction is
    /// scanned, the register that it defines is removed from the clique, and
    /// any register that it uses, if not already in the clique, is added to
    /// it. Notice that phi instructions do not contribute uses for clique
    /// construction. Only registers defined in the phi function are taken into
    /// consideration. The value sp_cut represents the highest virtual register
    /// that is not source of a spill operation. Spilled virtuals must not be
    /// alive in more than one clique.
    void build_list_cliques(
        unsigned sp_cut,
        std::vector<RegClique_Fer *> & cl,
        const MachineBasicBlock & mbb
    );


    /// Returns true if there is an edge linking r1 and r2, and false
    /// otherwise.
    bool is_connected(unsigned r1, unsigned r2);


    // This method scans the live range of v_reg, adding to the interference
    // graph all the interferences found between v_reg and any physical
    // register that is alive between the instructions cur and begin.
    void find_interferences(
        MachineBasicBlock::const_iterator cur,
        MachineBasicBlock::const_iterator & bgn,
        unsigned v_reg
    );


    /// This method sets the constraints on the mapping between physical
    /// registers and virtual registers. A constraint between two registers,
    /// e.g.: p_reg and v_reg is an indication that the live range of v_reg
    /// overlaps with the live range of a pre-colored register p_reg.
    void find_interferences(RegClique_Fer & ql, const MachineBasicBlock & mbb);


    /// This method removes all the interferences in the graph.
    void remove_interferences();

    private:

    /// Number of machine registers in the target architecture. It will never
    /// change during program execution.
    unsigned num_p_regs;


    /// This matrix holds the interferences between physical and virtual
    /// registers in the basic block. It avoids assigning a color to a
    /// virtual and later having to spill this virtual because this color
    /// is used by an overlapping physical (pre-colored) register.
    /// constraint_map[i] gives all the physical registers that interfere with
    /// the virtual i - MRegisterInfo::FirstVirtualRegister.
    std::vector< std::vector<bool> > constraint_map;


    /// This variable contains liveness analysis information.
    const EdgeLiveness_Fernando * edge_liveness;


    /// The bank of register keeps a reference to the machine function in
    /// order to get the architecture specific information.
    MachineFunction * machine_function;


    /// This method updates the set of registers with data from the given
    /// instruction. Any defined register is removed from the clique, and
    /// any used register is added to the clique. If the used register is
    /// already in the clique, nothing is done. The value sp_cut represents
    /// the highest virtual register that is not source of a spill operation.
    /// Spilled virtuals must not be alive in more than one clique.
    RegClique_Fer * update_list_of_cliques(
        unsigned sp_cut,
        const MachineBasicBlock & mbb,
        RegClique_Fer & q,
        const MachineInstr &mi
    );


    /// This method marks the physical register p_reg and the virtual
    /// register v_reg as overlapping. Thus, the register allocator must
    /// avoid assigning to v_reg the color assigned to v_reg.
    void add_interference(unsigned p_reg, unsigned v_reg);


    /// Free the memory, so the operating system can use it in other
    /// applications. It deletes the vector used to represent the interference
    /// graph.
    void releaseMemory();

    };

}

#endif
