//===-------------- KillingSites_Fer.cpp - Set of Registers ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it allows to store the killing sites of virtual registers.
//
// Date: Aug 10th, 2006
//
//===----------------------------------------------------------------------===////
// This class stores the killing sites of virtual registers. A killing site is
// the point where the register is last used. The killing sites of a basic
// block can be found quite quickly if liveness analysis has already been
// peformed. Start with the set of virtuals alive at the end of the basic
// block, and for each virtual, scan the basic block from end to beginning
// keeping the set of registers alive at each point. When an unseen register
// is found, mark that instruction as its killing site. Killing sites are
// store in a single vector. A map of instructions is used to index this
// vector.
//
//===----------------------------------------------------------------------===//

#ifndef KILLING_SITES_FER_H
#define KILLING_SITES_FER_H

#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/RegClique_Fer.h"

namespace llvm {

//===--------------------------------------------------------------------------
// This structure is used as a comparator in the map where instructions are
// stored. It compares the address of each instruction, in order to provide
// a relative ordering between them.
//===--------------------------------------------------------------------------
struct Cmp_Inst {
    bool operator()(const MachineInstr * mi1, const MachineInstr * mi2) const {
        return mi1 > mi2;
    }
};

class KillingSites_Fer {

    public:

    /// The constructor method. It receives a basic block and the clique of
    /// virtuals alive at the end of this basic block. I am passing the clique
    /// in the constructor in order to allow the basic block to be constant. If
    ///  the clique has to be computed internally, the basic block cannot be
    /// constant.
    KillingSites_Fer(const MachineBasicBlock * mbb, RegClique_Fer & ql);


    /// Destructor method. Erases the vector used to store killing sites, and
    /// the mapping used to index this vector.
    ~KillingSites_Fer() {
        this->i2index.clear();
        this->killed_regs.clear();
    }


    /// This method is used to indicate that old_i is a folded instruction, and
    /// new_i is the new instruction that represents it. All the registers
    /// killed in old_i will be considered killed in new_i.
    void fold_inst(const MachineInstr * old_i, const MachineInstr * new_i);


    /// This method inserts new_v as a virtual register killed at mi, if old_v
    /// is killed at mi, that is, it changes the virtual old_v by new_v in the
    /// killed list of mi.
    void replace_used_reg(const MachineInstr* i, unsigned newv, unsigned oldv);


    /// This method prints all killing sites found in the basic block used to
    /// create this object.
    void dump(const MachineFunction & mf);


    /// This method prints the vector used to store killing sites. It should be
    /// used for debugging purposes only, for the vector contains sequences of
    /// bites with very little meaning outside this object.
    void dump_vector();


    /// In order to understand this method, it is important to realize that this
    /// class has a simple notion of state. The state of instances of this class
    /// is given by the last instruction read. This method points the state to
    /// the given instruction, so that, next time the method has_dead_use() is
    /// called, it will be called for the instruction mi.
    void point_to_first_dead_use(const MachineInstr * mi);


    /// This method returns true if the current pointed instruction is the
    /// killing site of some virtual. To choose an instruction, use the method
    /// point_to_first_dead_use.
    bool has_dead_use() const;


    /// This method returns a virtual that is used at the current instruction,
    /// but is never subsequently used, and thus is dead. To choose an
    /// instruction, use the method point_to_first_dead_use.
    unsigned get_next_dead_use();


    /// In order to understand this method, it is important to realize that this
    /// class has a simple notion of state. The state of instances of this class
    /// is given by the last instruction read. This method points the state to
    /// the given instruction, so that, next time the method has_dead_def() is
    /// called, it will be called for the instruction mi.
    void point_to_first_dead_def(const MachineInstr * mi);


    /// This method returns true if the current pointed instruction contains
    /// dead definitions. To choose an instruction, use the method
    /// point_to_first_dead_def.
    bool has_dead_def() const;


    /// This method returns a virtual that is defined at the current
    /// instruction, but is never subsequently used, and thus is dead. To
    /// choose an instruction, use the method point_to_first_dead_def.
    unsigned get_next_dead_def();


    /// This method returns true if mi is the killing site of reg. Otherwise,
    /// it returns false.
    bool is_use_killing_site(const MachineInstr * mi, unsigned reg);


    /// Returns the basic block that is been scanned by this object. All the
    /// killing sites are instructions from the given basic block.
    const MachineBasicBlock * get_basic_block() const {
        return this->machine_bb;
    }

    private:

    /// This is the basic block that is been scanned by this object. All the
    /// killing sites are instructions from the given basic block.
    const MachineBasicBlock * machine_bb;


    /// This vector contains the killing information.
    std::vector< unsigned > killed_regs;


    /// This mapping is used to index the vector killed_regs. Each instruction
    /// is mapped to the position in the vector in which the killing
    /// killing information regarding that instruciton can be found.
    std::map< const MachineInstr *, unsigned, Cmp_Inst > i2index;


    /// This is used to keep the state of the object for the methods get/has
    /// dead_use/def.
    unsigned pointer;


    /// This is just an auxiliary function that is used to make sure that the
    /// value store in counter, and the size of the killing vector are the same.
    bool inc_counter(unsigned & counter);


    /// This method checks the registers used in mi and adds the ones not found
    /// in the clique ql as dead after mi, i.e, mi will be their killing site.
    void add_killed_uses(
        const MachineInstr * mi, 
        const RegClique_Fer & ql,
        unsigned & counter
    );


    /// This method checks the virtuals defined in mi, and marks mi as the
    /// killing site of those not used after it, i.e, not contained in ql.
    void add_dead_defs(
        const MachineInstr * mi,
        const RegClique_Fer & ql,
        unsigned & counter
    );


    /// Another auxiliary function that returns true if the register given as
    /// parameter is a physical register defined in the given instruction
    /// (explicitly or implicitly).
    bool is_defined_physical(unsigned reg, const MachineInstr * mi);

}; // class definition

} // namespace

#endif
