//===----------------- NeatPrint_Fer.cpp - Layout output ------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class contains some handy functions to generate strings describing
// instructions, register, etc.
//
// Date: June 17th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file contains functions that produce strings describing instructions,
// operands, and other data structures of the LLVM platform.
//
//===----------------------------------------------------------------------===//

#include <iostream>                              // cerr
#include <sstream>                               // ostringstream (RegClique)

#include "llvm/Function.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/RegMapping_Fer.h"

namespace llvm {

#ifndef NEAT_PRINTER_FER
#define NEAT_PRINTER_FER
 
class NeatPrint_Fer {

public:

//===--------------------------------------------------------------------------
/// Returns a string describing the name of the instruction.
//===--------------------------------------------------------------------------
static std::string opcode2string
                        (const MachineInstr & mi, const MachineFunction & mf) {
    const TargetMachine & tm = mf.getTarget();
    const TargetInstrInfo * tii = tm.getInstrInfo();
    const TargetInstrDescriptor & tid = tii->get(mi.getOpcode());
    std::ostringstream aux;
    aux << tid.Name;
    return aux.str();
}


//===--------------------------------------------------------------------------
/// Returns a string describing the name of the register. If it is a virtual
/// register, it will be printed in the format tn, e.g. t0, t1, ..., where n
/// is given by (reg - MRegisterInfo::FirstVirtualRegister). If it is physical
/// register, it will be printed in the format reg_name(reg), where reg_name is
/// the register's name as seen by the target architecture, e.g EAX, r1, etc.
//===--------------------------------------------------------------------------
static std::string op2string
                      (const unsigned reg, const MachineFunction & mf) {

    std::ostringstream reg_name;
    if(MRegisterInfo::isVirtualRegister(reg)) {
        reg_name << "t" << (reg - MRegisterInfo::FirstVirtualRegister);
    } else {
        const TargetMachine & target_machine = mf.getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        reg_name << reg_info->getName(reg);
    }
    reg_name << "(" << reg << ")";
    return reg_name.str();
}


//===--------------------------------------------------------------------------
/// Returns a string describing the name of the register. If it is a virtual
/// register, it will be printed in the format tn, e.g. t0, t1, ..., where n
/// is given by (reg - MRegisterInfo::FirstVirtualRegister). If it is physical
/// register, it will be printed in the format reg_name(reg), where reg_name is
/// the register's name as seen by the target architecture, e.g EAX, r1, etc.
/// The name of the operands is augmented with the name of the physical
/// register that has been assigned to it.
//===--------------------------------------------------------------------------
static std::string op2string(
    unsigned reg,
    const MachineFunction & mf,
    RegMapping_Fer & rm
) {
    std::ostringstream reg_name;
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    if(MRegisterInfo::isVirtualRegister(reg)) {
        reg_name << "t" << (reg - MRegisterInfo::FirstVirtualRegister);
        // Finding the name of the physical location:
        unsigned p_reg = rm.get_physical_location(reg);
        if(p_reg == RegMapping_Fer::NO_REG) {
            reg_name << "(spill)";
        } else {
            reg_name << "(" << reg_info->getName(p_reg) << ")";
        }
    } else {
        reg_name << reg_info->getName(reg);
    }

    return reg_name.str();
}


//===--------------------------------------------------------------------------
/// Returns a string describing the instruction opcode, plus all the operands
/// that are part of the instruction. The name of the operands is augmented
/// with the name of the physical register that has been assigned to it.
//===--------------------------------------------------------------------------
static std::string mi2string(
    const MachineInstr & mi,
    const MachineFunction & mf,
    RegMapping_Fer & rm
) {
    std::ostringstream mi_str;
    mi_str << NeatPrint_Fer::opcode2string(mi, mf);
    unsigned i = 0;
    while(i != mi.getNumOperands()) {
        const MachineOperand & mo = mi.getOperand(i);
        if (mo.isDef() && mo.isRegister() && mo.getReg()) {
            mi_str << " " << op2string(mo.getReg(), mf, rm);
        } else if(mo.isRegister() && mo.isUse() && mo.getReg()) {
            break;
        }
        i++;
    }
    mi_str << " := ";
    while(i != mi.getNumOperands()) {
        const MachineOperand & mo = mi.getOperand(i);
        if(mo.isRegister() && mo.getReg()) {
            mi_str << " " << op2string(mo.getReg(), mf, rm);
        }
        i++;
    }
    return mi_str.str();
}


//===--------------------------------------------------------------------------
/// Returns a string describing the instruction opcode, plus all the operands
/// that are part of the instruction.
//===--------------------------------------------------------------------------
static std::string mi2string
                        (const MachineInstr & mi, const MachineFunction & mf) {
    std::ostringstream mi_str;
    mi_str << NeatPrint_Fer::opcode2string(mi, mf);
    unsigned i = 0;
    while(i != mi.getNumOperands()) {
        const MachineOperand & mo = mi.getOperand(i);
        if (mo.isRegister() && mo.isDef() && mo.getReg()) {
            mi_str << " " << op2string(mo.getReg(), mf);
        } else if(mo.isRegister() && mo.isUse() && mo.getReg()) {
            break;
        }
        i++;
    }
    mi_str << " := ";
    while(i != mi.getNumOperands()) {
        const MachineOperand & mo = mi.getOperand(i);
        if(mo.isRegister() && mo.getReg()) {
            mi_str << " " << op2string(mo.getReg(), mf);
        }
        i++;
    }
    return mi_str.str();
}


//===--------------------------------------------------------------------------
/// This method prints all the instructions in the basic block, in a format
/// that helps debugging register allocation.
//===--------------------------------------------------------------------------
static std::string mbb2string(const MachineBasicBlock & mbb) {
    std::ostringstream bstr;
    bstr << NeatPrint_Fer::bbheader2string(mbb);
    const MachineFunction * mf = mbb.getParent();
    MachineBasicBlock::const_iterator mii = mbb.begin();
    while (mii != mbb.end()) {
        bstr << NeatPrint_Fer::mi2string(*mii, *mf) << "\n";
        mii++;
    }
    return bstr.str();
}


//===--------------------------------------------------------------------------
/// Returns a string describing the header of the machine basic block.
//===--------------------------------------------------------------------------
static std::string bbheader2string(const MachineBasicBlock & mbb) {
    std::ostringstream bstr;
    bstr << "Block (" << mbb.getNumber() << ")\n";
    return bstr.str();
}


//===--------------------------------------------------------------------------
/// Prints all the basic blocks that are part of the function
//===--------------------------------------------------------------------------
static void print_mf(const MachineFunction & mf) {
    for (MachineFunction::const_iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        std::cerr << mbb2string(*mbb) << "\n";
    }
}


//===--------------------------------------------------------------------------
/// This method prints the header information of the dot graph. This
/// information defines the font used for nodes and edges, plus the shape of
/// nodes. In this case, nodes have rectangular shape, and the font is
/// "Courier".
//===--------------------------------------------------------------------------
static void print_graph_header(const MachineFunction & mf) {
    // first prints he format of each node.
    const Function * f = mf.getFunction();
    std::cerr << "digraph " << f->getName() << " {\n\n";
    std::cerr << "node [\n";
    std::cerr << "  fontname = \"Courier\"\n";
    std::cerr << "  fontsize = \"10\"\n";
    std::cerr << "  shape = \"box\"\n";
    std::cerr << "]\n\n";
 
    // second prints he format of each edge.
    std::cerr << "edge [\n";
    std::cerr << "  fontname = \"Courier\"\n";
    std::cerr << "  fontsize = \"8\"\n";
    std::cerr << "]\n\n";
}


//===--------------------------------------------------------------------------
/// Prints the cfg of the given machine function using the dot format.
//===--------------------------------------------------------------------------
static void print_cfg(const MachineFunction & mf) {
    print_graph_header(mf);
    for (MachineFunction::const_iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        print_edges(*mbb);
        print_block_label(mf, *mbb);
    }
    std::cerr << "}\n";
}


//===--------------------------------------------------------------------------
/// Prints the cfg of the given machine function using the dot format. This
/// method will print the results of liveness analysis in the edges between
/// the basic blocks.
//===--------------------------------------------------------------------------
static void print_cfg(
    const MachineFunction & mf,
    const EdgeLiveness_Fernando & edge_liveness
) {
    print_graph_header(mf);
    for (MachineFunction::const_iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        print_edges(*mbb, edge_liveness);
        print_block_label(mf, *mbb);
    }
    std::cerr << "}\n";
}


//===--------------------------------------------------------------------------
/// Prints the cfg of the given machine function using the dot format. This
/// method will print the results of liveness analysis in the edges between
/// the basic blocks. Each virtual will be printed along with the physical
/// register to which it is mapped.
//===--------------------------------------------------------------------------
static void print_cfg(
    const MachineFunction & mf,
    const EdgeLiveness_Fernando & edge_liveness,
    RegMapping_Fer & rm
) {
    print_graph_header(mf);
    for (MachineFunction::const_iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        print_edges(mf, *mbb, edge_liveness, rm);
        print_block_label(mf, *mbb, rm);
    }
    std::cerr << "}\n";
}


//===--------------------------------------------------------------------------
/// This method prints all the edges from the block "MBB" to all its
/// successors. Edges are augmented with liveness analysis information.
/// The label of each edge consists on the set of registers that are
/// alive at that edge. This method will print the results of liveness
/// analysis in the edges. 
//===--------------------------------------------------------------------------
static void print_edges(
    const MachineFunction & mf,
    const MachineBasicBlock & mbb,
    const EdgeLiveness_Fernando & edge_liveness,
    RegMapping_Fer & rm
) {
    unsigned mbb_num = mbb.getNumber();
 
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    // Prints all the successors:
    for(MachineBasicBlock::const_succ_iterator s_it = mbb.succ_begin();
                                              s_it != mbb.succ_end(); s_it++) {
        const MachineBasicBlock* smbb = *s_it;
        unsigned suc_num = smbb->getNumber();
        std::string label;
        std::ostringstream laux;
        // prints the label:
        std::vector<unsigned> live_v;
        edge_liveness.to_vector(mbb, *smbb, live_v);
        laux << "[label = \"";
        for(unsigned u = 0; u < live_v.size(); u++) {
            unsigned p_reg = rm.get_physical_location(live_v[u]);
            unsigned v_reg = live_v[u] - MRegisterInfo::FirstVirtualRegister;
            laux << "t" << v_reg << "(" << reg_info->getName(p_reg) << ") ";
        }
        laux << "\"]";
        label = laux.str();
        std::cerr << "\"" << mbb_num << "\" -> " << "\"" << suc_num << "\""
                                                       << " " << label << "\n";
    }
}


//===--------------------------------------------------------------------------
/// This method prints all the edges from the block "MBB" to all its
/// successors. Edges are augmented with liveness analysis information.
/// The label of each edge consists on the set of registers that are
/// alive at that edge. This method will print the results of liveness
/// analysis in the edges.
//===--------------------------------------------------------------------------
static void print_edges(const MachineBasicBlock & mbb,
                                 const EdgeLiveness_Fernando & edge_liveness) {
    unsigned mbb_num = mbb.getNumber();
 
    // Prints all the successors:
    for(MachineBasicBlock::const_succ_iterator s_it = mbb.succ_begin();
                                              s_it != mbb.succ_end(); s_it++) {
        const MachineBasicBlock* smbb = *s_it;
        unsigned suc_num = smbb->getNumber();
        std::string label;
        std::ostringstream laux;
        laux << "[label = \"" << edge_liveness.to_string(mbb, *smbb) << "\"]";
        label = laux.str();
        std::cerr << "\"" << mbb_num << "\" -> " << "\"" << suc_num << "\""
                                                       << " " << label << "\n";
    }
}


//===--------------------------------------------------------------------------
/// Prints all the edges that leave the given basic block. The output format
/// is dot. The label of each basic block is its block number; thus, an edge
/// is like "n1" -> "n2", where n1 and n2 are integer non-negative numbers.
//===--------------------------------------------------------------------------
static void print_edges(const MachineBasicBlock & mbb) {
    unsigned mbb_num = mbb.getNumber();
    for(MachineBasicBlock::const_succ_iterator succ = mbb.succ_begin();
                                              succ != mbb.succ_end(); succ++) {
        const MachineBasicBlock* mbb_succ = *succ;
        unsigned succ_num = mbb_succ->getNumber();
        std::cerr << "\"" << mbb_num << "\" -> \"" << succ_num << "\"\n";
    }
}


//===--------------------------------------------------------------------------
/// Prints the label of the basic block in the control flow graph description.
/// The label of the basic block includes the block number, and all the
/// instructions that are in the block, in the sequence in which they are
/// executed.
//===--------------------------------------------------------------------------
static void print_block_label
                  (const MachineFunction & mf, const MachineBasicBlock & mbb) {
    unsigned block_number = mbb.getNumber();
    std::cerr << "\"" << block_number << "\"" << " [ label=\"* "
              << block_number << " *\\l";

    MachineBasicBlock::const_iterator mii = mbb.begin();
    while (mii != mbb.end()) {
        std::cerr << NeatPrint_Fer::mi2string(*mii, mf) << " \\l";
        mii++;
    }
    std::cerr << "\"]\n";

}

//===--------------------------------------------------------------------------
/// Prints the label of the basic block in the control flow graph description.
/// The label of the basic block includes the block number, and all the
/// instructions that are in the block, in the sequence in which they are
/// executed. This method augments the name of each virtual with the name of
/// the physical register that has been assigned to it.
//===--------------------------------------------------------------------------
static void print_block_label(
    const MachineFunction & mf,
    const MachineBasicBlock & mbb,
    RegMapping_Fer & rm
) {
    unsigned block_number = mbb.getNumber();
    std::cerr << "\"" << block_number << "\"" << " [ label=\"* "
              << block_number << " *\\l";

    MachineBasicBlock::const_iterator mii = mbb.begin();
    while (mii != mbb.end()) {
        std::cerr << NeatPrint_Fer::mi2string(*mii, mf, rm) << " \\l";
        mii++;
    }
    std::cerr << "\"]\n";

}

private:

};

#endif

}
