//===------------- PhiDeconstruction.h - Phi-elimination Pass -------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it eliminate phi functions after register allocation has been
// performed.
//
// Date: June 30th, 2006
//
//===----------------------------------------------------------------------===//
//
// Traditional algorithms used to destroy phi functions normally replace them
// by copy instructions. Although the classical algorithms preserve the
// semantics of the SSA-form program, they may increase the number of registers
// found during the register assignment phase. To avoid this problem, we
// replace phi functions by parallel copies. Parallel copies are normally
// implemented with xor instructions (Remember the trick of swaping two
// variables without an auxiliary using xor?). The register swap is a costly
// operation, because it demands three instructions to be implemented;
// therefore, it should be avoided whenever possible. One way of avoiding
// swaps is by means of move instructions. Basically, swap operations are not
// necessary in every time that phi instructions are destroyed. This happens
// when the definition vector contains machine registers not used in the
// parameter vector. This pass tries to use such optimization  before inserting
//  swap operations. It performs phi elimination in time linear in the size of
// the definition vector.
//
//===----------------------------------------------------------------------===//

#ifndef PHI_DECONSTRUCTION_FER_H
#define PHI_DECONSTRUCTION_FER_H

#include "llvm/fernando/PhiEq_Fer.h"

namespace llvm {

//===--------------------------------------------------------------------------
// This structure is used to store indexed values. It is useful to speed the
// phi elimination algorithm. There are situations in which the index of a
// value is needed. If this structure were not used, it would be necessary
// to scan the vector where the value is stored to discover its index.
//===--------------------------------------------------------------------------
struct ValueIndex {
    ValueIndex(unsigned v, unsigned i) {
        this->value = v;
        this->index = i;
    }
    unsigned value;
    unsigned index;
};


//===--------------------------------------------------------------------------
// This class implements an array of booleans, that in C++ can be stored
// as an array of bits. It is used as an auxiliary class to represent the
// "done" vector in the phi elimination algorithm.
//===--------------------------------------------------------------------------
class UnsignedVector_Fer {

    public:

    /// Creates a new vector of booleans with "size" cells, and fills it up
    /// with the elements in "src_vec".
    UnsignedVector_Fer(const std::vector<unsigned> & src_vec) {
        this->vec.resize(src_vec.size());
        for(unsigned u = 0; u < src_vec.size(); u++) {
            this->vec[u] = src_vec[u];
        }
        this->next = 0;
    }
    /// Destructor method: returns the memory to the operating system,
    /// where it belongs :)
    ~UnsignedVector_Fer() {
        this->vec.clear();
    }
    /// Returns the index of the given element in this vector. If the element
    /// is not present, it will return -1. 
    int get_index(unsigned u) {
        unsigned beginning = this->next;
        while(this->vec[this->next] != u) {
            this->next = (this->next + 1) % vec.size();
            if(this->next == beginning) {
                return -1;
            }
        }
        return this->next;
    }
    /// Returns the element in the given index
    unsigned get(unsigned u) {
        assert(u < this->vec.size() && "Index out of bound.");
        return this->vec[u];
    }
    /// Return true if element "u" is present in the vector, and false
    /// otherwise. 
    bool contains(unsigned u) {
        unsigned beginning = this->next;
        while(this->vec[this->next] != u) {
            this->next = (this->next + 1) % vec.size();
            if(this->next == beginning) {
                return false;
            }
        }
        return true;
    }
    /// Removes the element u from the set. If the element was present in the
    /// set, this method returns true. It returns false otherwise.
    bool remove_index(unsigned u) {
        assert(u < this->vec.size() && "Index out of bound.");
        if(this->vec[u] != RegMapping_Fer::NO_REG) {
            this->vec[u] = RegMapping_Fer::NO_REG;
            return true;
        }
        return false;
    }
    /// This method chooses a virtual from the vector and removes it.
    /// Which element is chosen should not be taken into consideration when
    /// invoking this method.
    unsigned choose_and_remove_any_virtual() {
        unsigned beginning = this->next;
        while(this->vec[this->next] == RegMapping_Fer::NO_REG ||
                             this->vec[this->next] == RegMapping_Fer::SP_REG) {
            this->next = (this->next + 1) % vec.size();
            if(this->next == beginning) {
                return RegMapping_Fer::NO_REG;
            }
        }
        unsigned ret_value = this->vec[this->next];
        this->vec[this->next] = RegMapping_Fer::NO_REG;
        return ret_value;
    }
    /// This method chooses a memory reference from the vector and removes it.
    /// Which element is chosen should not be taken into consideration when
    /// invoking this method. The returned value is the index of the removed
    /// element.
    unsigned choose_and_remove_any_mem_ref() {
        unsigned beginning = this->next;
        while(this->vec[this->next] != RegMapping_Fer::SP_REG) {
            this->next = (this->next + 1) % vec.size();
            if(this->next == beginning) {
                return RegMapping_Fer::NO_REG;
            }
        }
        this->vec[this->next] = RegMapping_Fer::NO_REG;
        return this->next;
    }
    /// Returns true if there are still virtual in the set; false otherwise.
    bool has_virtual() {
        unsigned beginning = this->next;
        while(this->vec[this->next] == RegMapping_Fer::NO_REG ||
                             this->vec[this->next] == RegMapping_Fer::SP_REG) {
            this->next = (this->next + 1) % vec.size();
            if(this->next == beginning) {
                return false;
            }
        }
        return true;
    }
    /// Returns true if there memory references in this set, false otherwise.
    bool has_memory_ref() {
        unsigned beginning = this->next;
        while(this->vec[this->next] != RegMapping_Fer::SP_REG) {
            this->next = (this->next + 1) % vec.size();
            if(this->next == beginning) {
                return false;
            }
        }
        return true;
    }
    /// This method fills reg_vec with the indices of the variables in phi_vec,
    /// according to this algorithm:
    /// Unsigned i := 0;
    /// while(i < size(phi_vec))
    /// |    Unsigned reg := phi_vec[i];
    /// |    reg_vec[reg] := i;
    /// |    i++;
    /// end while
    void fill_vector_indices(int size, std::vector<unsigned> & reg_vec) {
        reg_vec.clear();
        unsigned no_reg = RegMapping_Fer::NO_REG;
        reg_vec.assign(size, no_reg);
        unsigned u = 0;
        while(u < this->vec.size()) {
            unsigned reg = this->vec[u];
            if(reg != no_reg) {
                reg_vec[reg] = u;
            }
            u++;
        }
    }
    /// Returns the number of cells stored in this vector. Some of them may
    /// not contain valid registers. In this case, they would contain
    /// RegMapping_Fer::NO_REG or RegMapping_Fer::SP_REG
    unsigned size() {
        return this->vec.size();
    }

    private:

    std::vector<unsigned> vec;

    unsigned next;
};


/// TODO: I would like to declare this class as a LLVM pass. In this case I
/// would implement the getAnalysisUsage and getPassName functions. The problem
/// is that I don't know how to start this function after the register
/// allocator has finished his work. By now, I will implement the class as an
/// ordinary class, and I will name the main function runOnMachineFunction,
/// so it can be easily converted on a machine pass in the future.

  class PhiDeconstruction_Fer {

    public:

    /// Constructor method.
    PhiDeconstruction_Fer(
        EdgeLiveness_Fernando & el,
        RegMapping_Fer & rm,
        VirtRegMap & vrm
    ) {
        this->edge_liveness = & el;
        this->reg_mapping = & rm;
        this->vrm = & vrm;
    }

    /// runOnMachineFunction - Invokes the phi elimination algorithm. Normally
    /// this method would be private, and would be called by the pass manager,
    /// but in the current implementation, I don't really now how to
    /// implement this class as a pass.
    bool runOnMachineFunction(
        MachineFunction & machine_function
    );

    private:

    /// This variable denotes the number of physical registers in the target
    /// machine.
    unsigned num_p_regs;


    /// The bank of register keeps a reference to the machine function in
    /// order to get the architecture specific information.
    MachineFunction * machine_function;


    /// This object contains the liveness information for each edge of the
    /// control flow graph. It is necessary in order to discover which
    /// physical registers can be used for transfers between memory locations.
    EdgeLiveness_Fernando * edge_liveness;


    /// This object contains the mapping between spilled virtuals to stack
    /// slots.
    VirtRegMap * vrm;


    /// This object maps virtual registers to physical registers.
    RegMapping_Fer * reg_mapping;


    /// This vector marks which physical registers are been used, and thus
    /// cannot be used in implementing memory transfer sequences of
    /// instructions.
    std::vector<bool> occupied_physicals;

    /// This method removes all the occurrences of phi-functions from the
    /// given basic block. Before this method is called, instructions must be
    /// inserted in the control flow graph to reproduce the behavior of phi
    /// functions.
    void remove_phi_functions(MachineBasicBlock & mbb);


    /// This method inserts move and swap instructions in the control flow
    /// graph, in order to reproduce the semantics of phi-functions. Notice
    /// that the phi functions are not eliminated by this method.
    void insert_moves_and_swaps(MachineBasicBlock & mbb);


    /// This method determines which virtual registers are reaching this basic
    /// block alive. These virtuals are alive in the edge that goes from src
    /// basic block to dst basic block.
    void set_occupied_regs(MachineBasicBlock & org, MachineBasicBlock & dst);


    /// This method determines which virtual registers are reaching this basic
    /// block alive. Let L be the set of live registers: L = live_edge(org,
    /// dst) - live_use + live_def.
    void set_occupied_regs(
        MachineBasicBlock & org,
        MachineBasicBlock & dst,
        UnsignedVector_Fer & def,
        UnsignedVector_Fer & use
    );


    /// This method returns a register that can be assigned to the given
    /// virtual register. It will return RegMapping_Fer::NO_REG if there
    /// is no register available.
    unsigned get_free_register(unsigned v_reg);


    /// This method determines if p_reg is not currently allocated to any
    /// virtual register alive in the beginning of the given basic block.
    bool is_register_available(unsigned p_reg);


    /// This method inserts swap instructions in order to destroy phi-functions.
    /// A swap instruction is implemented with xor instructions, and each swap
    /// will produce three xor's. Therefore, move instructions should be
    /// inserted whenever possible; however, in (a, b) := (b, a), move
    /// instructions cannot be used without demanding an extra register.
    void insert_swap_instructions(
        MachineBasicBlock & mbb,
        std::vector<ValueIndex> & dif,
        UnsignedVector_Fer & done_def,
        UnsignedVector_Fer & done_use,
        PhiEq_Fer & phi
    );


    /// This method helps destroying phi functions by using move instructions
    /// whenever applicable. Move instructions are preferable instead of swaps
    /// because generally they can be implemented cheaply. If def vector
    /// contains virtual v, which is not in the use vector, it is always safe
    /// to add a move into v. The basic block mbb is the block where the move
    /// instruction should be inserted.
    void insert_move_instructions(
        MachineBasicBlock & mbb,
        std::vector<ValueIndex> & dif,
        UnsignedVector_Fer & done_def,
        UnsignedVector_Fer & done_use,
        PhiEq_Fer & ph
    );


    /// This method store virtuals in the use set of the phi function when the
    /// corresponding definition is a spilled value. This method is always safe,
    /// and can be called any time.
    void resolve_virt2mem(
        UnsignedVector_Fer & done_def,
        UnsignedVector_Fer & done_use,
        MachineBasicBlock & origin,
        PhiEq_Fer & phi_eq
    );


    /// This method inserts instructions to transfer a value in memory to
    /// another memory location.
    bool allocate_memory_transfers(
        UnsignedVector_Fer & done_def,
        UnsignedVector_Fer & done_use,
        MachineBasicBlock & origin,
        PhiEq_Fer & phi_eq
    ); 


    /// The method assumes that use_v and def_v have been mapped to memory. It
    /// inserts instructions to tranfer the contents of use_v's location to
    /// def_v's location.
    bool insert_memory_transfer(
        MachineBasicBlock & mbb,
        unsigned src,
        unsigned dst
    );


    /// This method inserts a store instruction at the end of the basic block
    /// mbb. This instruction stores the virtual use_v to the stack position
    /// assigned to the virtual def_v.
    void insert_store(
        MachineBasicBlock & mbb,
        unsigned use_v,
        unsigned def_v
    );


    /// This method fills vector r with indices that contain valid registers in
    /// v1 and do not contain valid registers in v2. That is, this method fills
    /// vector r with the registers that are not used in v2, and that are used
    /// in v1. It implements set intersection in linear time.
    void colors_v1_minus_colors_v2 (
        UnsignedVector_Fer & v1,
        UnsignedVector_Fer & v2,
        std::vector<ValueIndex> & r
    );


    /// This method inserts a sequence of instructions at the end of basic
    /// block mbb in order to swap the values of registers r1 and r2. Which
    /// instructions are inserted dependes on the target architecture.
    /// Normally swaps are implemented with three xor instructions.
    void add_swap(MachineBasicBlock & mbb, unsigned r1, unsigned r2);


    /// This method inserts a move instruction at the end of basic block mbb.
    /// The move instruction copies the value of register src to register dst.
    /// If the source of the move is a spill, then, instead of inserting a move
    /// instruction, this method directly loads the spilled value into the
    /// destiny.
    void add_move(
        MachineBasicBlock & mbb,
        unsigned src,
        unsigned dst
    );

  }; // end of class PhiDeconstruction_Fer

} // end llvm namespace

#endif
