//===-------------- FernandoPrint.cpp - Printing prototype ----------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and is a prototype intended to help the author to better
// understand LLVM.
//
// Date: April 3rd, 2006
//
//===----------------------------------------------------------------------===//
//
// This file prints all the instructions in the given machine function.
//
//===----------------------------------------------------------------------===//

#include <vector>
#include "llvm/CodeGen/Passes.h"               // FernandoPassID
#include "llvm/Type.h"
#include "llvm/Instructions.h"                 // MachineOperand, etc
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/CodeGen/LiveVariables.h"
#include "llvm/BasicBlock.h"
#include "llvm/CodeGen/MachineFunctionPass.h"  // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/ADT/DepthFirstIterator.h"       // df_ext_iterator
#include "llvm/fernando/LiveSet_Fernando.h"
#include "VirtRegMap.h"
#include <iostream>                            // cerr
#include <sstream>                             // ostringstream
#include "llvm/fernando/PhiEq_Fer.h"

using namespace llvm;

namespace {

  #define INST_LEN 20
  #define OPRT_LEN 10

  class FernandoPrint : public MachineFunctionPass {

    // This class contains the variables currently alive in the basic
    // block. It is a type of set.
    Fernando_LiveSet *liveVariables;

    // This vector holds true for each value that represents a virtual
    // register, and false otherwise;
    std::vector<bool> VirtRegTable;

    // This class computes live variable information for
    // each virtual and register allocatable physical register in a function.
    LiveVariables *LV;

    /// All target-specific information can be accessed through this interface.
    const TargetMachine *TM;

    /// Information about the bank of registers implemented by the actual
    /// machine can be accessed through this interface.
    const MRegisterInfo *RegInfo;

    /// This variable will contain information about the loop structure of
    /// the control flow graph.
    const LoopInfo *loopInfo;

    public:
      virtual void getAnalysisUsage(AnalysisUsage &AU) const {
        AU.addRequired<LiveVariables>();
        AU.addRequired<LoopInfo>();
        AU.setPreservesAll();
      }

      virtual const char *getPassName() const {
        return "Print machine instructions - Fernando.";
      }

    private:
      /// runOnMachineFunction - Prints information about each basic block.
      bool runOnMachineFunction(MachineFunction &Fn);

      /// AllocateBasicBlock - For the given basic block, prints the name
      /// of the basic block, and information about each of the
      /// instructions.
      void PrintBasicBlock(MachineBasicBlock &MBB);

      /// Prints the registers implicitly read and defined by the instruction
      void PrintImplicit(const TargetInstrDescriptor &TID);

      /// Prints the virtual registers used and defined by the instruction.
      void PrintOperands(MachineInstr *MI);

      /// Prints the name of the register.
      int PrintRegister(MachineOperand &MO);

      /// Prints the name, number, and neat header for the basic block.
      void PrintBBHeader(MachineBasicBlock &MBB);

      /// This method prints informations about each PHI node:
      void PrintPHINodes(MachineBasicBlock &MBB);

      /// Prints the list of virtual registers:
      void PrintVirtRegs();

      /// Prints the definition site, the killing instructions, and all the
      /// basic blocks whose live range of the register crosses.
      void PrintLiveInterval(unsigned reg);

      /// Fills str with n characters like 'c', where n = size(str) + packSize.
      /// It returns the number of characters printed.
      int PrintPackedStr(const char * str, int packSize, char c);
      int PrintPackedStr(std::string str, int packSize, char c);

      /// print the register alive in the in set of the first block of the
      /// function.
      void printLiveInRegs(MachineFunction &Fn);

      /// print the registers alive at the end of the basic block.
      void printLiveOutRegs(MachineFunction &Fn);

      /// Gives the name of the basic block
      std::string getBBName(MachineBasicBlock &MBB) {
        const BasicBlock *BB = MBB.getBasicBlock();
        std::ostringstream aux;
        std::string bbName;
        aux << BB->getName();
        return aux.str();
      }
  };

  RegisterPass<FernandoPrint> Z("allocpty", "Print machine instructions.");

}


// the external declaration of FernandoPassID is in llvm/CodeGen/Passes.h
const PassInfo *llvm::FernandoPassID = Z.getPassInfo();


void FernandoPrint::PrintPHINodes(MachineBasicBlock &MBB) {

    PhiEq_Fer phi_mtx(MBB);
    std::cerr << phi_mtx.to_string();

}


void FernandoPrint::PrintLiveInterval(unsigned reg) {
  if(MRegisterInfo::isVirtualRegister(reg)) {
    LiveVariables::VarInfo vi = LV->getVarInfo(reg);
    const TargetInstrInfo &TII = *TM->getInstrInfo();
    std::cerr << "t" << (reg - MRegisterInfo::FirstVirtualRegister);
    // Definition site --------------------------------
    {
      MachineInstr* defInst = vi.DefInst;
      MachineBasicBlock* mbb = defInst->getParent();
      std::cerr << "\n  Def: " << mbb->getNumber();
    }
    // Live Blocks ------------------------------------
    {
      bool isFirst = true;
      for (unsigned i = 0, e = vi.AliveBlocks.size(); i != e; ++i) {
        if (vi.AliveBlocks[i]) {
          if(isFirst) {
            std::cerr << "\n  Alive in blocks: " << i;
            isFirst = false;
          } else {
            std::cerr << ", " << i;
          }
        }
      }
    }
    // Killing sites   --------------------------------
    {
      std::vector<MachineInstr*> killingSites = vi.Kills;
      std::cerr << "\n  Uses:";
      if (!killingSites.empty()) {
        bool isFirst = true;
        for (unsigned i = 0, e = killingSites.size(); i != e; ++i) {
          MachineBasicBlock* mbb = killingSites[i]->getParent();
          if(isFirst) {
            std::cerr << mbb->getNumber();
            isFirst = false;
          }
          else {
            std::cerr << ", " << mbb->getNumber();
          }
        }
      }
      std::cerr << "\n";
    }
  }
}


void FernandoPrint::PrintVirtRegs() {
  unsigned virtReg = 0;

  std::cerr << "Printing virtual registers.\n";

  for(virtReg = 0; virtReg < VirtRegTable.size(); virtReg++) {
    if(VirtRegTable[virtReg]) {
      PrintLiveInterval(virtReg);
    }
  }
  std::cerr << "\n";
}


void FernandoPrint::PrintBasicBlock(MachineBasicBlock &MBB) {
  PrintBBHeader(MBB);
  PrintPHINodes(MBB);
/*
  // loop over each instruction
  MachineBasicBlock::iterator MII = MBB.begin();
  const TargetInstrInfo &TII = *TM->getInstrInfo();
  while (MII != MBB.end()) {
    MachineInstr *MI = MII++;
    if(MI->getOpcode() == TargetInstrInfo::PHI) {
      std::cerr << "PHI instruction:\n";
    }
    const TargetInstrDescriptor &TID = TII.get(MI->getOpcode());
    PrintPackedStr(TID.Name, INST_LEN, ' ');
    PrintOperands(MI);
    PrintImplicit(TID);
    std::cerr << "\n" << liveVariables->to_string() << "\n";
    std::cerr << "\n";
  }
*/
}


int FernandoPrint::PrintPackedStr(std::string str, int packSize, char c) {
  int strSize = str.size();
  int incSize = strSize < packSize ? packSize - strSize : 0 ;
  std::string strAux(incSize, c);
  std::cerr << str << strAux;
  return (strSize > packSize) ? strSize : packSize ;
}


int FernandoPrint::PrintPackedStr(const char * str, int packSize, char c) {
  int strSize = strlen(str);
  int incSize = strSize < packSize ? packSize - strSize : 0 ;
  std::string strAux(incSize, c);
  std::cerr << str << strAux;
  return (strSize > packSize) ? strSize : packSize ;
}


void FernandoPrint::PrintBBHeader(MachineBasicBlock &MBB) {
  int i = 0;
  const BasicBlock *BB = MBB.getBasicBlock();
  std::string headerLine(79, '=');
  std::ostringstream aux;
  std::string blockName;
  unsigned loopDepth = loopInfo->getLoopDepth(BB);
  aux << MBB.getNumber() << ") " << BB->getName() << " ";
  blockName = aux.str();
  for(i = blockName.size(); i < 79; i++) {
    blockName += "-";
  }
  std::cerr << ", Loop depth = " << loopDepth << "\n";
  std::cerr << headerLine << "\n";
  for(MachineBasicBlock::succ_iterator succ = MBB.succ_begin();
      succ != MBB.succ_end();
      succ++) {
    MachineBasicBlock* mbb = *succ;
    const BasicBlock *bb = mbb->getBasicBlock();
    std::cerr << " -> " << bb->getName() << "\n";
  }
  std::cerr << blockName << "\n";
}


int FernandoPrint::PrintRegister(MachineOperand &MO) {
  int returnValue = 0;
  std::ostringstream aux;
  std::string result;
  if (MO.getReg()) {
    unsigned regNum = MO.getReg();
    if(MRegisterInfo::isPhysicalRegister(regNum)) {
      aux << RegInfo->getName(regNum) << "(" << "r" << regNum << ")";
    }
    else {
      aux << "t" << (regNum - MRegisterInfo::FirstVirtualRegister);
    }
    if(regNum > 0) {
      if(VirtRegTable.size() < regNum) {
        VirtRegTable.resize(regNum + 1);
      }
      VirtRegTable[regNum] = true;
    }
  }
  result = aux.str();
  returnValue = PrintPackedStr(result, OPRT_LEN, '-');
  return returnValue;
}


void FernandoPrint::PrintOperands(MachineInstr *MI) {
  bool isFirst = true;
  int column = INST_LEN;
  int colChange = 0;

  // first, print all the definitions
  for (unsigned i = 0; i != MI->getNumOperands(); ++i) {
    MachineOperand& MO = MI->getOperand(i);
    if (MO.isDef() && MO.isRegister() && MO.getReg()) {
      if(isFirst) {
        colChange += PrintRegister(MO);
        isFirst = false;
      } else {
        std::cerr << ", ";
        colChange += PrintRegister(MO);
      }
    }
  }
  if(colChange == 0) {
    PrintPackedStr("", OPRT_LEN, '-');
  }

  // next, print all the uses
  std::cerr << " := ";
  isFirst = true;
  for (unsigned i = 0; i != MI->getNumOperands(); ++i) {
    MachineOperand& MO = MI->getOperand(i);
    if (MO.isUse() && MO.isRegister() && MO.getReg()) {
      if(isFirst) {
        PrintRegister(MO);
        isFirst = false;
      } else {
        std::cerr << ", ";
        PrintRegister(MO);
      }
    }
  }

  liveVariables->update_live_ranges(MI);
}


void FernandoPrint::PrintImplicit(const TargetInstrDescriptor &TID) {
  bool isFirst = true;
  // print all the implicit uses:
  for (const unsigned *ImplicitUses = TID.ImplicitUses;
       *ImplicitUses; ++ImplicitUses) {
    if(isFirst) {
      std::cerr << "Imp. Uses: r" << *ImplicitUses;
      isFirst = false;
    } else {
      std::cerr << ", r" << *ImplicitUses;
    }
  }
  isFirst = true;
  // print all the implicit defs:
  for (const unsigned *ImplicitDefs = TID.ImplicitDefs;
       *ImplicitDefs; ++ImplicitDefs) {
    if(isFirst) {
      std::cerr << "Imp. Defs: r" << *ImplicitDefs;
      isFirst = false;
    } else {
      std::cerr << ", r" << *ImplicitDefs;
    }
  }
  std::cerr << "\n";
}


void FernandoPrint::printLiveInRegs(MachineFunction &Fn) {
   // Printing live in registers.
   bool hasLiveIn = false;
   for (MachineFunction::livein_iterator I = Fn.livein_begin(),
          E = Fn.livein_end(); I != E; ++I) {
     assert(MRegisterInfo::isPhysicalRegister(I->first) &&
            "Cannot have a live-in virtual register!");
       std::cerr << "live in: " << I->first << "\n";
       hasLiveIn = true;
   }
    if(!hasLiveIn) {
        std::cerr << "The function does not contain registers live in\n";
    }
}


void FernandoPrint::printLiveOutRegs (MachineFunction &Fn) {
    for (MachineFunction::liveout_iterator I = Fn.liveout_begin(),
         E = Fn.liveout_end(); I != E; ++I) {
         assert(MRegisterInfo::isPhysicalRegister(*I) &&
         "Cannot have a live-in virtual register!");
         std::cerr << "Register in live out: " << *I << "\n";
    }
}


bool FernandoPrint::runOnMachineFunction(MachineFunction &Fn) {

  this->TM = &Fn.getTarget();
  this->RegInfo = TM->getRegisterInfo();
  this->LV = &getAnalysis<LiveVariables>();
  this->loopInfo = &getAnalysis<LoopInfo>();
  this-> liveVariables = new Fernando_LiveSet(LV);
  VirtRegTable.assign(((*&Fn).getSSARegMap())->getLastVirtReg(), false);

/*
  printLiveInRegs(Fn);

  {
    MachineBasicBlock *Entry = Fn.begin();
    std::set<MachineBasicBlock*> Visited;

    for (df_ext_iterator<MachineBasicBlock*> DFI = df_ext_begin(Entry, Visited),
       E = df_ext_end(Entry, Visited); DFI != E; ++DFI) {
      MachineBasicBlock *MBB = *DFI;
      PrintBasicBlock(*MBB);
    }
  }

  printLiveOutRegs(Fn);

  PrintVirtRegs();


  std::cerr << "\n---------------------------------\n";
  Fn.dump();
*/
  Fn.viewCFG();

  VirtRegTable.clear();
  delete liveVariables;

  return false;
}
