//===----------- IG_Builder_Fer - Builds the interference graph ----------===//
//
//                     The LLVM Compiler Infrastructure
//
// Given a control flow graph, this pass builds the correcsponding interference
// graph. This file has been implemented by fernando at cs dot ucla dot edu.
//
// Date: July 10th, 2006
//
//===---------------------------------------------------------------------===//
//
// This pass builds the interference graph for a target control flow graph.
// Two nodes in the interference graph are connected if, and only if, their
// live ranges overlap in the control flow graph.
// Important: this version of the interference graph does not contains the
// full interfering pairs. It only holds interferences between physical
// registers and virtual registers. This helps in the implementation of the
// register allocation pass, and avoids using too much memory and time to
// hold the interference graph. Actually, this graph uses about one fourth
// of the memory that it would use if the full interference graph were to be
// built and stored.
//
//===---------------------------------------------------------------------===//

#include <iostream>                               // cerr

#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Instructions.h"                   // MachineOperand
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/IG_Builder_Fer.h"

using namespace llvm;

static RegisterAnalysis<IG_Builder_Fer>
X("interference_graph", "Interference graph builder");


//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::releaseMemory() {
    this->constraint_map.clear();
}


//===--------------------------------------------------------------------------
// Returns true if there is an edge linking p_reg and v_reg, and false
// otherwise.
//===--------------------------------------------------------------------------
bool IG_Builder_Fer::is_connected(unsigned p_reg, unsigned v_reg) {
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                           "Target register must be physical");
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                            "Source register must be virtual");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->constraint_map.size()) {
        return false;
    } else if(this->constraint_map[base_reg].size() == 0) {
        return false;
    } else {
        return this->constraint_map[base_reg][p_reg];
    }
}


//===--------------------------------------------------------------------------
// This method prints the constraints in the standard error output.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::dump_constraints() {
    for(unsigned t = 0; t < this->constraint_map.size(); t++) {
        std::cerr << "(" << t << ")" << ": ";
        for(unsigned u = 0; u < this->constraint_map[t].size(); u++) {
            if( this->constraint_map[t][u] ) {
                std::cerr <<
                  NeatPrint_Fer::op2string(u, * this->machine_function) << " ";
            }
        }
        std::cerr << "\n";
    }
}


//===--------------------------------------------------------------------------
// This method removes all the interferences in the graph.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::remove_interferences() {
    for(unsigned t = 0; t < constraint_map.size(); t++) {
        for(unsigned u = 0; u < constraint_map[t].size(); u++) {
            constraint_map[t][u] = false;
        }
    }
}


//===--------------------------------------------------------------------------
// This method marks the physical register p_reg and the virtual register
// v_reg as overlapping. Thus, the register allocator must avoid assigning to
// v_reg the color assigned to v_reg.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::add_interference(unsigned p_reg, unsigned v_reg) {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
//if(base_reg == 30 && (p_reg == 73 || p_reg == 138)) exit(1);
    if(base_reg >= this->constraint_map.size()) {
        this->constraint_map.resize(base_reg + 1);
    }
    if(this->constraint_map[base_reg].size() == 0) {
        this->constraint_map[base_reg].resize(this->num_p_regs);
    }
    this->constraint_map[base_reg][p_reg] = true;
    // Add interferences for all the physical alias of p_reg
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        this->constraint_map[base_reg][*a] = true;
    }
}



//===--------------------------------------------------------------------------
// This method scans the live range of v_reg, adding to the interference graph
// all the interferences found between v_reg and any physical register that is
// alive between the instructions cur and begin.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::find_interferences(
    MachineBasicBlock::const_iterator cur,
    MachineBasicBlock::const_iterator & bgn,
    unsigned v_reg
) {
    const TargetMachine & target_machine = machine_function->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const MachineInstr * mi;

    while(true) {
        mi = cur;
        const TargetInstrDescriptor & tid = tii.get(mi->getOpcode());

        // compute interferences against implicitly used regs
        if(tid.ImplicitUses) {
            const unsigned *implicit_use = tid.ImplicitUses;
            while(*implicit_use) {
                add_interference(*implicit_use, v_reg);
            }
        }

        // compute interferences against explicitly used regs
        for(unsigned u = 0; u < mi->getNumOperands(); u++) {
            const MachineOperand & mo = mi->getOperand(u);
            if(mo.isRegister() && mo.isUse() && mo.getReg()) {
                if(MRegisterInfo::isPhysicalRegister(mo.getReg())) {
                    add_interference(mo.getReg(), v_reg);
                }
            }
        }

        // if this is the first instruction in the basic block, it is over.
        if(cur == bgn) {
            return;
        } else {
            cur--;
        }

        mi = cur;

        // compute interferences against implicitly defined regs
        if(tid.ImplicitDefs) {
            for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
                add_interference(*implicit_def, v_reg);
            }
        }

        // compute interferences against implicitly defined regs
        bool reg_is_defined_in_mi = false;
        for(unsigned u = 0; u < mi->getNumOperands(); u++) {
            const MachineOperand & mo = mi->getOperand(u);
            if(mo.isRegister() && mo.isDef() && mo.getReg()) {
                if(MRegisterInfo::isPhysicalRegister(mo.getReg())) {
                    add_interference(mo.getReg(), v_reg);
                } else {
                    if(mo.getReg() == v_reg) {
                        reg_is_defined_in_mi = true;
                    }
                }
            }
        }

        // If the v_reg is defined in mi, we are done.
        if(reg_is_defined_in_mi) {
            return;
        }
    }
}


//===--------------------------------------------------------------------------
// This method sets the constraints on the mapping between physical registers
// and virtual registers. A constraint between two registers, e.g.: p_reg and
// v_reg is an indication that the live range of v_reg overlaps with the live
// range of a pre-colored register p_reg.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::find_interferences
                          (RegClique_Fer & ql, const MachineBasicBlock & mbb) {

    MachineBasicBlock::const_iterator mbb_cur = mbb.end();
    MachineBasicBlock::const_iterator mbb_bgn = mbb.begin();
    if(mbb_cur == mbb_bgn) {
        return;
    } else {
        mbb_cur--;
    }

    const MachineFunction * mf = mbb.getParent();
    const SSARegMap * srm = mf->getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    for(
        unsigned v_reg = MRegisterInfo::FirstVirtualRegister;
        v_reg < last_v_reg;
        v_reg++
    ) {
        if(ql.contains_reg(v_reg)) {
            find_interferences(mbb_cur, mbb_bgn, v_reg);
        }
    }

    // remember to add a set to avoid double checking the same operand
    mbb_cur = mbb.end();
    while (mbb_cur != mbb_bgn) {
        mbb_cur--;
        const MachineInstr * mi = mbb_cur;
        for(unsigned u = 0; u < mi->getNumOperands(); u++) {
            const MachineOperand & mo = mi->getOperand(u);
            if(mo.isRegister() && mo.getReg() && mo.isUse()) {
                if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                    if(!ql.contains_reg(mo.getReg())) {
                        find_interferences(mbb_cur, mbb_bgn, mo.getReg());
                        ql.add_reg(mo.getReg());
                    }
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method modifies the cliques in the cl vector to reflect the newly
// created virtual registers. Such virtuals have been created due to spilling.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::rebuild_list_cliques
(unsigned sp, std::vector<RegClique_Fer *> &cl, const MachineBasicBlock &mbb) {

    // Augment each clique, so they can support the newly created virtual
    // registers.
    const MachineFunction *mf = mbb.getParent();
    SSARegMap * srm = mf->getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    for(unsigned u = 0; u < cl.size(); u++) {
        cl[u]->grow_virtual(last_v_reg - MRegisterInfo::FirstVirtualRegister);
    }

    MachineBasicBlock::const_iterator mbbi = mbb.end();
    const MachineInstr * mi;
    unsigned clique_counter = 1;

    while (mbbi != mbb.begin()) {
        mbbi--;
        mi = mbbi;
        bool isPhi = (mi->getOpcode() == TargetInstrInfo::PHI) ? true : false ;

        // Remove the virtual registers that only represent spills:
        RegClique_Fer * clique = cl[clique_counter];
        std::vector<unsigned> contents;
        clique->to_vector(contents);
        for(unsigned i = 0; i < contents.size(); i++) {
            if(contents[i] > sp) {
                clique->remove_reg(contents[i]);
            }
        }

        // add the new uses (some of the old uses will be added too, but it
        // does not matter).
        for(unsigned i = 0; i != mi->getNumOperands(); i++) {
            const MachineOperand & mo = mi->getOperand(i);
            if (mo.isUse() && mo.isRegister() && mo.getReg() && (!isPhi) ) {
                clique->add_reg(mo.getReg());
            }
        }

        clique_counter++;
    }
}


//===--------------------------------------------------------------------------
// This method updates the set of registers with data from the given
// instruction. Any defined register is removed from the clique, and any used
// register is added to the clique. If the used register is already in the
// clique, nothing is done. The value sp_cut represents the highest virtual
// register that is not source of a spill operation. Spilled virtuals must not
// be alive in more than one clique.
//===--------------------------------------------------------------------------
RegClique_Fer * IG_Builder_Fer::update_list_of_cliques(
    unsigned sp_cut,
    const MachineBasicBlock & mbb,
    RegClique_Fer & q,
    const MachineInstr & mi
) {
    RegClique_Fer * ret_value = new RegClique_Fer(q);
    ret_value->remove_definitions(mi);
    // Remove the values that represent spilled registers. They have very
    // short live ranges, only one clique wide.
    std::vector<unsigned> regs;
    ret_value->to_vector(regs);
    for(unsigned u = 0; u < regs.size(); u++) {
        if(regs[u] > sp_cut) {
            ret_value->remove_reg(regs[u]);
        }
    }
    ret_value->add_uses(mi);
    return ret_value;
}


//===--------------------------------------------------------------------------
// If liveness analysis information is available, the set of cliques in a basic
// block can be found with a backwards linear scan. The first clique will be
// the out set of the basic block. Each time a new instruction is scanned, the
// register that it defines is removed from the clique, and any register that
// it uses, if not already in the clique, is added to it. Notice that phi
// instructions do not contribute uses for clique construction. Only registers
// defined in the phi function are taken into consideration. The value sp_cut
// represents the highest virtual register that is not source of a spill
// operation. Spilled virtuals must not be alive in more than one clique.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::build_list_cliques(
    unsigned sp_cut,
    std::vector<RegClique_Fer *> & cl,
    const MachineBasicBlock & mbb
) {
    RegClique_Fer * q = new RegClique_Fer
                           (*this->machine_function, mbb, this->edge_liveness);
    cl.push_back( q );

    MachineBasicBlock::const_iterator mbbi = mbb.end();
    const MachineInstr * mi;
    unsigned i = 0;
    while (mbbi != mbb.begin()) {
        mbbi--;
        mi = mbbi;
        cl.push_back( update_list_of_cliques(sp_cut, mbb, *cl[i], *mi) );
        i++;
    }
}


//===--------------------------------------------------------------------------
// This method builds the interference graph between physical and virtual
// registers.
//===--------------------------------------------------------------------------
bool IG_Builder_Fer::runOnMachineFunction(MachineFunction &mf) {

    this->machine_function = & mf;

    // initializes internal data structures:
    // First, finds the number of physical registers:
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    this->num_p_regs = reg_info->getNumRegs();

    // secondly, finds the number of virtual registers:
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;
    this->constraint_map.resize(num_v_regs + 1);

    // gets the liveness analysis information
    this->edge_liveness = & getAnalysis<EdgeLiveness_Fernando>();

    for (MachineFunction::iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        RegClique_Fer ql(mf, *mbb, this->edge_liveness);
        find_interferences(ql, *mbb);
    }

    return false;

}
