//===-------------- KillingSites_Fer.cpp - Set of Registers ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it allows to store the killing sites of virtual registers.
//
// Date: Aug 10th, 2006
//
//===----------------------------------------------------------------------===////
// This class stores the killing sites of virtual registers. A killing site is
// the point where the register is last used. The killing sites of a basic
// block can be found quite quickly if liveness analysis has already been
// peformed. Start with the set of virtuals alive at the end of the basic
// block, and for each virtual, scan the basic block from end to beginning
// keeping the set of registers alive at each point. When an unseen register
// is found, mark that instruction as its killing site. Killing sites are
// store in a single vector. A map of instructions is used to index this
// vector.
//
//===----------------------------------------------------------------------===//

#include <map>
#include <set>
#include <vector>
#include <iostream>

#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"           // It is necessary in NeatP..

#include "llvm/fernando/NeatPrint_Fer.h"

#include "llvm/fernando/KillingSites_Fer.h"

using namespace llvm;

#define DEAD_USES 65535
#define DEAD_DEFS 65534

//===--------------------------------------------------------------------------
// This is just an auxiliary function that is used to make sure that the value
// store in counter, and the size of the killing vector are the same.
//===--------------------------------------------------------------------------
bool KillingSites_Fer::inc_counter(unsigned & counter) {
    counter++;
    if(this->killed_regs.size() <= counter) {
        this->killed_regs.resize(2*(this->killed_regs.size() + 1), DEAD_USES);
        return true;
    }
    return false;
}


//===--------------------------------------------------------------------------
// Another auxiliary function that returns true if the register given as
// parameter is a physical register defined in the given instruction
// (explicitly or implicitly).
//===--------------------------------------------------------------------------
bool KillingSites_Fer::is_defined_physical
                                      (unsigned reg, const MachineInstr * mi) {
    if(MRegisterInfo::isVirtualRegister(reg)) {
        return false;
    }

    // Verifies if the register is implicitly defined in this instruction: 
    const MachineFunction * mf = this->machine_bb->getParent();
    const TargetMachine & target_machine = mf->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi->getOpcode());
    if(tid.ImplicitDefs) {
        for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
            if(*implicit_def == reg) {
                return true;
            }
        }
    }

    // Verifies if the physical register is explicitly defined:
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() && mo.isDef()) {
            if(mo.getReg() == reg) {
                return true;
            }
        }
    }

    return false;
}


//===--------------------------------------------------------------------------
// This method checks the registers used in mi and adds the ones not found in
// the clique ql as dead after mi, i.e, mi will be their killing site.
//===--------------------------------------------------------------------------
void KillingSites_Fer::add_killed_uses(
    const MachineInstr * mi,
    const RegClique_Fer & ql,
    unsigned & counter
) {
    std::set<unsigned> seen_regs;
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() && mo.isUse()) {
            unsigned reg = mo.getReg();
            if(!ql.contains_reg(reg) || is_defined_physical(reg, mi)) {
                if(seen_regs.find(reg) != seen_regs.end()) {
                    continue;
                } else {
                    seen_regs.insert(reg);
                    this->killed_regs[counter] = reg;
                    inc_counter(counter);
                }
            }
        }
    }
    // Add the implicit dead uses to the list:
    const MachineFunction * mf = this->machine_bb->getParent();
    const TargetMachine & target_machine = mf->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi->getOpcode());
    if(tid.ImplicitUses) {
        for (const unsigned *implicit_use = tid.ImplicitUses;
                                               *implicit_use; ++implicit_use) {
            unsigned p_reg = *implicit_use;
            if( ! ql.contains_reg(p_reg) ) {
                this->killed_regs[counter] = p_reg;
                inc_counter(counter);
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method checks the virtuals defined in mi, and marks mi as the killing
// site of those not used after it, i.e, not contained in ql.
//===--------------------------------------------------------------------------
void KillingSites_Fer::add_dead_defs(
    const MachineInstr * mi,
    const RegClique_Fer & ql,
    unsigned & counter
) {
    // Add the explicit dead definitions to the list:
    std::set<unsigned> seen_regs;
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() && mo.isDef()) {
            unsigned reg = mo.getReg();
            if(!ql.contains_reg(reg)) {
                if(seen_regs.find(reg) != seen_regs.end()) {
                    continue;
                } else {
                    seen_regs.insert(reg);
                    this->killed_regs[counter] = reg;
                }
                inc_counter(counter);
            }
        }
    }
    // Add the implicit dead definitions to the list:
    const MachineFunction * mf = this->machine_bb->getParent();
    const TargetMachine & target_machine = mf->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi->getOpcode());
    if(tid.ImplicitDefs) {
        for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
            unsigned p_reg = *implicit_def;
            if( ! ql.contains_reg(p_reg) ) {
                this->killed_regs[counter] = p_reg;
                inc_counter(counter);
            }
        }
    }
}


//===--------------------------------------------------------------------------
// The constructor method. It receives a basic block and the clique of virtuals
// alive at the end of this basic block. I am passing the clique in the
// constructor in order to allow the basic block to be constant. If the clique
// has to be computed internally, the basic block cannot be constant.
//===--------------------------------------------------------------------------
KillingSites_Fer::KillingSites_Fer(
    const MachineBasicBlock * mbb,
    RegClique_Fer & ql
) {
    this->machine_bb = mbb;

    // I am assigning an arbitrary size here, to avoid resizing the vector
    // all the time. If this is not enough, I will add more space to the vector
    // while it is been filled.
    this->killed_regs.assign(2 * mbb->size(), DEAD_USES);

    MachineBasicBlock::const_iterator mbbi = mbb->end();
    unsigned counter = 0;
    while (mbbi != mbb->begin()) {
        mbbi--;
        const MachineInstr * mi = mbbi;
        this->i2index[mi] = counter;
        add_killed_uses(mi, ql, counter);
        this->killed_regs[counter] = DEAD_DEFS;
        inc_counter(counter);
        add_dead_defs(mi, ql, counter);
        inc_counter(counter);
        ql.remove_definitions(*mi);
        ql.add_uses(*mi);
    }
}


//===--------------------------------------------------------------------------
// This method is used to indicate that old_i is a folded instruction, and
// new_i is the new instruction that represents it. All the registers killed
// in old_i will be considered killed in new_i.
//===--------------------------------------------------------------------------
void KillingSites_Fer::fold_inst(
    const MachineInstr * old_i,
    const MachineInstr * new_i
) {
    assert(this->i2index.find(old_i) != this->i2index.end() &&
                                 "Machine instruction is not a killing site.");
    unsigned index = this->i2index[old_i];
    this->i2index[new_i] = index;
}


//===--------------------------------------------------------------------------
// This method inserts new_v as a virtual register killed at mi, if old_v is
// killed at mi, that is, it changes the virtual old_v by new_v in the killed
// list of mi.
//===--------------------------------------------------------------------------
void KillingSites_Fer::replace_used_reg(
    const MachineInstr * mi,
    unsigned new_v,
    unsigned old_v
) {
    assert(this->i2index.find(mi) != this->i2index.end() &&
                                "Machine instruction is not in killing list.");
    unsigned index = this->i2index[mi];
    while(this->killed_regs[index] != DEAD_DEFS) {
        if(this->killed_regs[index] == old_v) {
            this->killed_regs[index] = new_v;
        }
        index++;
    }
}


//===--------------------------------------------------------------------------
// In order to understand this method, it is important to realize that this
// class has a simple notion of state. The state of instances of this class is
// given by the last instruction read. This method points the state to the
// given instruction, so that, next time the method has_dead_def() is called,
// it will be called for the instruction mi.
//===--------------------------------------------------------------------------
void KillingSites_Fer::point_to_first_dead_def(const MachineInstr * mi) {
    assert(this->i2index.find(mi) != this->i2index.end() &&
                                "Machine instruction is not in killing list.");
    this->pointer = this->i2index[mi];
    while(this->killed_regs[this->pointer] != DEAD_DEFS) {
        this->pointer++;
    }
    this->pointer++;
}


//===--------------------------------------------------------------------------
// This method returns true if the current pointed instruction contains dead
// definitions. To choose an instruction, use the method
// point_to_first_dead_def.
//===--------------------------------------------------------------------------
bool KillingSites_Fer::has_dead_def() const {
    return this->killed_regs[this->pointer] != DEAD_USES;
}


//===--------------------------------------------------------------------------
// This method returns a virtual that is defined at the current instruction,
// but is never subsequently used, and thus is dead. To choose an instruction,
// use the method point_to_first_dead_def.
//===--------------------------------------------------------------------------
unsigned KillingSites_Fer::get_next_dead_def() {
    assert(this->killed_regs[this->pointer] != DEAD_USES &&
                    "There is not more dead defs for the current instruction");
    unsigned dead_def = this->killed_regs[this->pointer];
    this->pointer++;
    return dead_def;
}


//===--------------------------------------------------------------------------
// In order to understand this method, it is important to realize that this
// class has a simple notion of state. The state of instances of this class is
// given by the last instruction read. This method points the state to the
// given instruction, so that, next time the method has_dead_use() is called,
// it will be called for the instruction mi.
//===--------------------------------------------------------------------------
void KillingSites_Fer::point_to_first_dead_use(const MachineInstr * mi) {
    assert(this->i2index.find(mi) != this->i2index.end() &&
                                "Machine instruction is not in killing list.");
    this->pointer = this->i2index[mi];
}


//===--------------------------------------------------------------------------
// This method returns true if the current pointed instruction is the killing
// site of some virtual. To choose an instruction, use the method
// point_to_first_dead_use.
//===--------------------------------------------------------------------------
bool KillingSites_Fer::has_dead_use() const {
    return this->killed_regs[this->pointer] != DEAD_DEFS;
}


//===--------------------------------------------------------------------------
// This method returns a virtual that is used at the current instruction,
// but is never subsequently used, and thus is dead. To choose an instruction,
// use the method point_to_first_dead_use.
//===--------------------------------------------------------------------------
unsigned KillingSites_Fer::get_next_dead_use() {
    assert(this->killed_regs[this->pointer] != DEAD_DEFS &&
                    "There is not more dead uses for the current instruction");
    unsigned dead_use = this->killed_regs[this->pointer];
    this->pointer++;
    return dead_use;
}


//===--------------------------------------------------------------------------
// This method returns true if mi is the killing site of reg. Otherwise, it
// returns false.
//===--------------------------------------------------------------------------
bool KillingSites_Fer::is_use_killing_site
                                      (const MachineInstr * mi, unsigned reg) {
    point_to_first_dead_use(mi);
    while(has_dead_use()) {
        if(this->killed_regs[this->pointer] == reg) {
            return true;
        } else {
            this->pointer++;
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// This method prints all killing sites found in the basic block used to
// create this object.
//===--------------------------------------------------------------------------
void KillingSites_Fer::dump(const MachineFunction & mf) {
    std::map< const MachineInstr *, unsigned, Cmp_Inst >::iterator pair =
                                                         this->i2index.begin();
    while(pair != this->i2index.end()) {
        const MachineInstr * mi = (*pair).first;
        unsigned index = (*pair).second;
        std::cerr << NeatPrint_Fer::opcode2string(*mi, mf) << ") Dead uses: ";
        while(this->killed_regs[index] != DEAD_USES) {
            if(this->killed_regs[index] == DEAD_DEFS) {
                std::cerr << ", Dead defs: ";
            } else {
                std::cerr << " " <<
                        NeatPrint_Fer::op2string(this->killed_regs[index], mf);
            }
            index++;
        }
        std::cerr << "\n";
        pair++;
    }
}


//===--------------------------------------------------------------------------
// This method prints the vector used to store killing sites. It should be used
// for debugging purposes only, for the vector contains sequences of bites with
// very little meaning outside this object.
//===--------------------------------------------------------------------------
void KillingSites_Fer::dump_vector() {
    for(unsigned u = 0; u < this->killed_regs.size(); u++) {
        std::cerr << " " << this->killed_regs[u];
    }
    std::cerr << "\n";
}
