//===----------- FernandoLiveStack.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it holds the variables alive at certain point during a scan of
// the control flow graph.
//
// Date: May 26th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file can be used during a scan of the control flow graph to hold
// the values of variables alive at each point of the graph traversal.
// It is essentially a stack of live variables. Every time a variable is
// defined, it can be added to the stack. When it is killed, it can be
// removed. This file only handle virtual registers, because physical
// registers are not supposed to be alive across basic blocks.
//
//===----------------------------------------------------------------------===//

#include <vector>
#include <sstream>                          // ostringstream
#include <iostream>                         // cerr

#include "llvm/Instructions.h"              // MachineOperand
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/CodeGen/LiveVariables.h"

#include "llvm/fernando/LiveSet_Fernando.h" // class definition

using namespace llvm;

//===--------------------------------------------------------------------------
// This function scans all the operands of target instruction, and mark them
// as alive. Because the first operands are always definition, it will stop
// after the first use operand is found.
//===--------------------------------------------------------------------------
void Fernando_LiveSet::mark_defs_alive(MachineInstr *mi) {
    for (unsigned i = 0; i != mi->getNumOperands(); ++i) {
        MachineOperand &MO = mi->getOperand(i);
        if (MO.isDef() && MO.isRegister() && MO.getReg()) {
            unsigned reg = MO.getReg();
            if(MRegisterInfo::isVirtualRegister(reg)) {
                unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
                if (live_variables.size() <= base_reg) {
                    live_variables.resize(base_reg+1);
                }
                live_variables[base_reg] = true;
            }
        } else {
            break;
        }
    }
}


//===--------------------------------------------------------------------------
// This function scans all the operands in the target instruction that are
// uses, and mark them as dead. This function is used because the killing
// list in LiveVariables::VarInfo does not contain the uses of PHI functions,
// so, it should be called only for PHI instructions.
//===--------------------------------------------------------------------------
void Fernando_LiveSet::mark_uses_dead(MachineInstr *mi) {
    for (unsigned i = 0; i != mi->getNumOperands(); ++i) {
        MachineOperand &MO = mi->getOperand(i);
        if (MO.isUse() && MO.isRegister() && MO.getReg()) {
            unsigned reg = MO.getReg();
            if(MRegisterInfo::isVirtualRegister(reg)) {
                unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
                assert(live_variables.size() > base_reg &&
                       "Fer_LiveSet: Index out of range in mark uses dead.");
                live_variables[base_reg] = false;
            }
        }
    }
}


//===--------------------------------------------------------------------------
// Returns a string representation of the live temporaries. This information
// will be printed in a format like "(t1, t2, ..., tn)", where ti is one of
// the virtual registers alive.
//===--------------------------------------------------------------------------
std::string Fernando_LiveSet::to_string() {
    bool isFirst = true;
    std::ostringstream aux;
    aux << "(";
    for (unsigned i = 0, e = live_variables.size(); i != e; ++i) {
        if(live_variables[i]) {
            if(isFirst) {
                isFirst = false;
                aux << "t" << i;
            } else {
                aux << ", " << "t" << i;
            }
        }
    }
    aux << ")";
    return aux.str();
}


//===--------------------------------------------------------------------------
// if this instruction defines a register, then marks the register as
// alive. If it uses registers, and the registers are not used
// thereafter, it marks the registers as dead.
// Param MO: the operand. It will be considered only if it encodes a virtual
// register.
//===--------------------------------------------------------------------------
void Fernando_LiveSet::update_live_ranges(MachineInstr *mi) {

    // handle declaration.
    mark_defs_alive(mi);

    // handle the last uses.
    if (mi->getOpcode() == TargetInstrInfo::PHI) {
        mark_uses_dead(mi);
    } else {
        for ( LiveVariables::killed_iterator ki = live_ranges->killed_begin(mi),
              ke = live_ranges->killed_end(mi);
              ki != ke; ++ki)
        {
            unsigned reg = *ki;
            if (MRegisterInfo::isVirtualRegister(reg)) {
                unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
                assert(live_variables.size() > base_reg &&
                       "Fer_LiveSet: Index out of range in vector access.");
                live_variables[base_reg] = false;
            }
        }
    }
}
