//===-------- RegMapping_Fer.cpp - Bank of Machine Registers ---------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the bank of machine registers for register allocation
// purposes.
//
// Date: May 27th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class represents the bank of machine registers for register allocation
// purposes. Essentially, it marks which physical registers are being used,
// and which ones are free to use. For each physical register in use, it keeps
// track of which virtual registers they use.
//
//===----------------------------------------------------------------------===//

#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/RegClique_Fer.h"
#include "llvm/fernando/IG_Builder_Fer.h"

#include "llvm/fernando/RegMapping_Fer.h"

using namespace llvm;


//===--------------------------------------------------------------------------
// Constructor method. Initializes the data structures. The machine function
// parameter is only used to provide initializing information, such as the
// number of virtual and physical registers.
//===--------------------------------------------------------------------------
RegMapping_Fer::RegMapping_Fer(MachineFunction & mf, IG_Builder_Fer & ig) {
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    unsigned num_p_regs = reg_info->getNumRegs();
    // initialized the list of physical registers beeing used. The declaration
    // of this array can be found in llvm/CodeGen/MachineFunction.h.
    this->p_regs_ever_used = new bool[num_p_regs];
    std::fill(this->p_regs_ever_used, this->p_regs_ever_used+num_p_regs,false);
    mf.setUsedPhysRegs(this->p_regs_ever_used);
    // TODO verify why RegMapping_Fer::NO_REG cannot be used here...
    this->p2v.assign(num_p_regs, 65535);
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    this->last_register_non_spill = srm->getLastVirtReg();
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;
    this->v2p.assign(num_v_regs, 65535);
    this->machine_function = & mf;
    this->ig = & ig;
}


//===--------------------------------------------------------------------------
// Constructor method. It produces a mapping that is a copy of the set
// initially given as parameter.
//===--------------------------------------------------------------------------
RegMapping_Fer::RegMapping_Fer(RegMapping_Fer & rm) {
    this->p2v.assign(rm.p2v.begin(), rm.p2v.end());
    this->v2p.assign(rm.v2p.begin(), rm.v2p.end());
    this->machine_function = rm.machine_function;
    this->p_regs_ever_used = rm.p_regs_ever_used;
    this->ig = rm.ig;
    this->last_register_non_spill = rm.last_register_non_spill;
}


//===--------------------------------------------------------------------------
// Increase the number of virtual registers stored in the object by one. This
// method returns the new number of virtual registers.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::grow() {
    this->v2p.resize(this->v2p.size() + 1);
    this->v2p[this->v2p.size() - 1] = RegMapping_Fer::NO_REG;
    return this->v2p.size();
}


//===--------------------------------------------------------------------------
// Destructor method. Gives back to the operating system the memory that was in
// use during the object's execution.
//===--------------------------------------------------------------------------
RegMapping_Fer::~RegMapping_Fer() {
    this->p2v.clear();
    this->v2p.clear();
    // The p_regs_ever_used cannot be deleted, because it will be used in
    // other passes.
}


//===--------------------------------------------------------------------------
// This method determines if p_reg is not currently allocated to any virtual
// register. In order to verify this, it does not suffice to check if the
// cell of p_reg in the physical register table contains the NO_REG value,
// because physical registers contains alias. For instance, in the X86
// architecture, the physical registers AL, AX and EAX are all alias, and if
// one of them is in use, the others are not free. If p_reg is not been
// used to allocate any virtual, the method returns true. It returns false
// otherwise.
//===--------------------------------------------------------------------------
bool RegMapping_Fer::is_register_available(unsigned p_reg) {

    if (this->p2v[p_reg] != RegMapping_Fer::NO_REG) {
        return false;
    }

    const TargetMachine & target_machine = this->machine_function->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        if (this->p2v[*a] != RegMapping_Fer::NO_REG) {
            return false;
        }
    }

    return true;
}


//===--------------------------------------------------------------------------
// This method returns the virtual register that was assigned to the given
// physical register. If the physical register does not contain any virtual
// register, the method will return RegMapping_Fer::NO_REG.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_assigned_data(unsigned p_reg) {
    assert(p_reg < this->p2v.size() && "Invalid physical register.");
    unsigned data = this->p2v[p_reg];
    if(data == RegMapping_Fer::NO_REG) {
        const TargetMachine & target_machine = machine_function->getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            if (this->p2v[*a] != RegMapping_Fer::NO_REG) {
                data = this->p2v[*a];
                break;
            }
        }
    }
    return data;
}


//===--------------------------------------------------------------------------
// This method returns the physical register that was assigned to the virtual
// register. If the virtual register has not been allocated to any register, it
// will return null. It is also possible that the given parameter describes a
// physical register. In this case, it is a pre-colored register, and its
// location must be the physical register itself.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_physical_location(unsigned reg) {
    if(MRegisterInfo::isVirtualRegister(reg)) {
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        assert(base_reg < this->v2p.size() && "Invalid virtual register.");
        return this->v2p[base_reg];
    } else {
        return this->p2v[reg];
    }
}


//===--------------------------------------------------------------------------
// This method verifies if p_reg can be used to allocated v_reg. This will be
// the case if both registers belong to the same class.
//===--------------------------------------------------------------------------
bool RegMapping_Fer::compatible_class(unsigned v_reg, unsigned p_reg) {
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                           "Target register must be physical");
    const TargetRegisterClass *trc =
                    this->machine_function->getSSARegMap()->getRegClass(v_reg);
    return trc->contains(p_reg);
}


//===--------------------------------------------------------------------------
// This method returns a register that can be assigned to the given
// virtual register. It will return RegMapping_Fer::NO_REG if there
// is no register available. Notice that is is possible that there will
// be a register available, and even though this method returns NO_REG.
// This will happen if all the available registers are pre-colored
// registers that interfere with this one is some point of the control
// flow graph.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_free_register(unsigned v_reg) {
    const TargetRegisterClass *trc =
                    this->machine_function->getSSARegMap()->getRegClass(v_reg);
    TargetRegisterClass::iterator trci =
                         trc->allocation_order_begin(* this->machine_function);
    TargetRegisterClass::iterator trce =
                           trc->allocation_order_end(* this->machine_function);

    for (; trci != trce; ++trci) {
        if (is_register_available(*trci)) {
            if( ! this->ig->is_connected(*trci, v_reg) ) {
                return *trci;
            }
        }
    }

    return RegMapping_Fer::NO_REG;
}


//===--------------------------------------------------------------------------
// This method is very similar to get_free_register. Thid different is
// that here even a pre-colored register that interferes with v_reg can
// be returned.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_unused_reg(unsigned v_reg) {
    const TargetRegisterClass *trc =
                    this->machine_function->getSSARegMap()->getRegClass(v_reg);
    TargetRegisterClass::iterator trci =
                         trc->allocation_order_begin(* this->machine_function);
    TargetRegisterClass::iterator trce =
                           trc->allocation_order_end(* this->machine_function);

    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    for (; trci != trce; ++trci) {
        // notice that we are not testing to see if there is an interference
        // between the chosen register and the given parameter.
        if (is_register_available(*trci)) {
            return *trci;
        }
    }

    return RegMapping_Fer::NO_REG;
}


//===--------------------------------------------------------------------------
// This method clears the machine register that is assigned to the virtual
// v_reg. It has the effect of spilling v_reg, and after it is called, any
// register from the same class as v_reg will be able to be allocated. That is,
// after liberate_color is called, allocate_color will return true if the
// parameters are from the same class of registers.
//===--------------------------------------------------------------------------
void RegMapping_Fer::liberate_color(unsigned reg) {
    if(MRegisterInfo::isVirtualRegister(reg)) {
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        unsigned p_reg = this->v2p[base_reg];
        if(p_reg != RegMapping_Fer::NO_REG) {
            if(this->p2v[p_reg] == reg) {
                this->p2v[p_reg] = RegMapping_Fer::NO_REG;
            }
        }
    } else {
        this->p2v[reg] = RegMapping_Fer::NO_REG;
    }
}


//===--------------------------------------------------------------------------
// This method does the same thing as the method allocate_color, but it can
// only be used with spilled registers.
//===--------------------------------------------------------------------------
void RegMapping_Fer::set_color_spilled_register(unsigned vr, unsigned pr) {

    // The target register must be virtual, and must have been created due to
    // a spill:
    assert(MRegisterInfo::isVirtualRegister(vr)
                                    && "The target register must be virtual.");

    // The location must be a physical register, BUT IT DOES NOT HAVE TO BE
    // FREE. Remember that we are allocating only spilled registers.
    assert(MRegisterInfo::isPhysicalRegister(pr)
                             && "Target location is not a physical register.");
    unsigned base_reg = vr - MRegisterInfo::FirstVirtualRegister;
    assert(compatible_class(vr, pr) && "Incompatible register class.");
    this->p2v[pr] = vr;
    this->v2p[base_reg] = pr;
    this->p_regs_ever_used[pr] = true;
}


//===--------------------------------------------------------------------------
// This method allocated reg into p_reg. It performs a number of tests, if any
// of them fails, an exception will be raised. It checks if p_reg is a
// physical register, and if it is free. If reg is virtual, it verifies if
// it already has a color. It would be an error to assign a register twice
// because in SSA-form each variable is defined only once, and it verifies if
// reg and p_reg are compatible, that is, they belong to the same class.
//===--------------------------------------------------------------------------
void RegMapping_Fer::allocate_color(unsigned reg, unsigned p_reg) {
    // The location must be a physical register, and it must be free.
    assert(MRegisterInfo::isPhysicalRegister(p_reg)
                             && "Target location is not a physical register.");
    assert((this->p2v[p_reg] == RegMapping_Fer::NO_REG)
                                          && "Physical register is not free.");

    bool is_v = MRegisterInfo::isVirtualRegister(reg);
    bool is_p = MRegisterInfo::isPhysicalRegister(reg);
    assert( (is_v || is_p) && "Unknow register specification.");

    if(is_v) {
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        assert(compatible_class(reg, p_reg) && "Incompatible register class.");
        this->p2v[p_reg] = reg;
        this->v2p[base_reg] = p_reg;
        this->p_regs_ever_used[p_reg] = true;
    } else if(is_p) {
        assert(reg == p_reg
                 && "Attempt to assign incorrect color to physical register.");
        this->p2v[reg] = p_reg;
    }
}


//===--------------------------------------------------------------------------
// This method cleans the mapping between physical registers and virtual
// registers, and if the registers in the given clique have already been
// allocated, new values will be assigned to the bank of machine registers to
// reflect the allocation. This method is generally called before starting the
// allocation of a basic block. The sp value is used to distinguish spilled
// virtuals from non-spilled virtuals. It avoids that spilled virtuals in the
// first clique of the vector be added as alive in the input set of the basic
// block. The clique ql contains the set of registers that are alive in the
// starting point of the basic block.
//===--------------------------------------------------------------------------
void RegMapping_Fer::reset_physical_mapping(unsigned sp, RegClique_Fer & ql) {
    for(unsigned u = 0; u < this->p2v.size(); u++) {
        this->p2v[u] = RegMapping_Fer::NO_REG;
    }
    std::vector<unsigned> new_alloc;
    ql.to_vector(new_alloc);
    for(unsigned u = 0; u < new_alloc.size(); u++) {
        unsigned reg = new_alloc[u];
        if(MRegisterInfo::isVirtualRegister(reg)) {
            if(reg <= sp) {
                unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
                if(this->v2p[base_reg] != RegMapping_Fer::NO_REG) {
                    this->p2v[this->v2p[base_reg]] = reg;
                    this->p_regs_ever_used[this->v2p[base_reg]] = true;
                }
            }
        } else {
            this->p2v[reg] = reg;
            this->p_regs_ever_used[reg] = true;
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns a string in the format "t1(r1), t2(r2), ... tm(rm),
// where virtual registers ti are allocated in machine registers ri.
//===--------------------------------------------------------------------------
std::string RegMapping_Fer::virtual_to_string() {
    std::ostringstream aux;
    bool is_first = true;
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for(unsigned i = 0; i < this->v2p.size(); i++) {
        if(this->v2p[i] != RegMapping_Fer::NO_REG) {
            const char * phys_reg_name = reg_info->getName(this->v2p[i]);
            if(is_first) {
                aux << "t" << i << "(" << phys_reg_name << ")";
                is_first = false;
            } else {
                aux << ", " << "t" << i << "(" << phys_reg_name << ")";
            }
        }
    }
    return aux.str();
}


//===--------------------------------------------------------------------------
// This method returns a string in the format "r1(t1), r2(t2), ... rm(tm),
// where physical registers ri is holding the value of virtual register ti.
//===--------------------------------------------------------------------------
std::string RegMapping_Fer::physical_to_string() {
    std::ostringstream aux;
    bool is_first = true;
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for(unsigned i = 0; i < this->p2v.size(); i++) {
        if(this->p2v[i] != RegMapping_Fer::NO_REG) {
            unsigned reg = this->p2v[i];
            std::ostringstream reg_name;
            reg_name << NeatPrint_Fer::op2string(reg, * this->machine_function);
            std::ostringstream phys_name;
            phys_name << NeatPrint_Fer::op2string(i, * this->machine_function);
            std::ostringstream num_name;
            if(is_first) {
                aux << phys_name.str() << "[" << reg_name.str() << "]";
                is_first = false;
            } else {
                aux << ", " << phys_name.str() << "[" << reg_name.str() << "]";
            }
        }
    }
    return aux.str();
}
