//===----------- ThreeAddr2TwoAddr_Fer.cpp - Printing prototype -----------===//
//
//             Register Allocation via Coloring of Chordal Graphs 
//
// This pass converts a three address instruction to a two address instruction.
//
// Date: May 28th, 2006
//
//===----------------------------------------------------------------------===//
//
// In LLVM, all the instructions are three address code, but some target
// architectures have two address instructions. Therefore, it is necessary to
// make this conversion, what is exaclty what this class does.
//
//===----------------------------------------------------------------------===//

#include <iostream>                              // cerr

#include "llvm/CodeGen/Passes.h"                 // EdgeLivenessFernandoPassID
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Instructions.h"                   // MachineOperand
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include <llvm/Target/TargetMachine.h>
#include "llvm/ADT/DepthFirstIterator.h"         // df_ext_iterator

#include "llvm/fernando/ThreeAddr2TwoAddr_Fer.h" // class definition

using namespace llvm;


//===--------------------------------------------------------------------------
// Finds the two address instructions in the basic block and converts them to
// two address instructions. We assume that, if the instruction is in the
// format a := b + a, then it is commutative. We also assume that the first
// three operands of the instruction are registers.
//===--------------------------------------------------------------------------
bool ThreeAddr2TwoAddr_Fer::remove3address(MachineBasicBlock & mbb) {
    bool made_change = false;
    for (MachineBasicBlock::iterator mii = mbb.begin(), mei = mbb.end();
                                                           mii != mei; ++mii) {
        MachineInstr * mi = mii;
        unsigned opcode = mi->getOpcode();
        if(!this->tii->isTwoAddrInstr(opcode)) {
            continue;
        }

        // Now we know that we are dealing with a two address instruction.
        // Start by checking that this instruction is in the expected format:
        // a := b + c
        assert(mi->getOperand(0).isRegister() && mi->getOperand(0).getReg() &&
             mi->getOperand(0).isDef() && "two address instruction invalid");
        assert(mi->getOperand(1).isRegister() && mi->getOperand(1).getReg() &&
             mi->getOperand(1).isUse() && "two address instruction invalid");
        assert(mi->getOperand(2).isRegister() && mi->getOperand(2).getReg() &&
             mi->getOperand(2).isUse() && "two address instruction invalid");

        unsigned reg_a = mi->getOperand(0).getReg();
        unsigned reg_b = mi->getOperand(1).getReg();
        unsigned reg_c = mi->getOperand(2).getReg();

        if (reg_a != reg_b) {
            if(reg_a == reg_c) {
                // a := b + a -> a := a + b
                assert(this->tii->isCommutableInstr(opcode) &&
                                              "Instruction is not commutable");
                MachineInstr * new_mi = this->tii->commuteInstruction(mi);
                assert(new_mi && "Failure while commuting instruction");
                if(mi != new_mi) {
                    mbb.insert(mii, new_mi);
                    mbb.erase(mii);
                    mi = new_mi;
                }
            } else {
                // a := b + c -> a := b, a := a + c
                const TargetRegisterClass* rc =
                                  this->mf->getSSARegMap()->getRegClass(reg_a);
                // Inserting a := b
                this->ri->copyRegToReg(mbb, mii, reg_a, reg_b, rc);
                // replace all occurences of b with a
                for (unsigned i = 1, e = mi->getNumOperands(); i != e; ++i) {
                    if (mi->getOperand(i).isRegister() &&
                                         mi->getOperand(i).getReg() == reg_b) {
                        mi->getOperand(i).setReg(reg_a);
                    }
                }
            }
        }

        // After the previous transformations, we have a := a + x. We can
        // remove the second operator, and change the first into a use.
        assert(mi->getOperand(0).isDef() && "Wrong instruction format");
        mi->getOperand(0).setUse();
        mi->RemoveOperand(1);
        made_change = true;
    }
    return made_change;
}


//===--------------------------------------------------------------------------
// Checkes each instruction in the function, and converts the three address
// instructions to two address if necessary.
//===--------------------------------------------------------------------------
bool ThreeAddr2TwoAddr_Fer::runOnMachineFunction(MachineFunction & mf) {

    this->mf = & mf;
    const TargetMachine & tm = mf.getTarget();
    this->ri = tm.getRegisterInfo();
    this->tii = tm.getInstrInfo();

    // traverse the control flow graph in depth first order pinpointing the
    // registers alive in each edge:
    for (MachineFunction::iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        remove3address(*mbb);
    }

    return false;
}
